from copy import copy
import logging
from typing import Dict, Iterable, List, Mapping, Set, Tuple

from ymmsl import (
        Component, Configuration, Model, MPICoresResReq, MPINodesResReq,
        Operator, Reference, ResourceRequirements, ThreadedResReq)

from libmuscle.planner.resources import OnNodeResources, Resources
from libmuscle.util import instance_indices


_logger = logging.getLogger(__name__)


_PredSuccType = Dict[Component, Set[Tuple[Component, int]]]


class ModelGraph:
    """Represents a yMMSL model as a graph."""
    def __init__(self, model: Model) -> None:
        """Create a ModelGraph.

        This is essentially a helper class that makes it easier to
        analyse a yMMSL model definition.

        Args:
            model: The model to represent.
        """
        self._model = model

        self._direct_superpreds: _PredSuccType = {}
        self._direct_predecessors: _PredSuccType = {}
        self._direct_subpreds: _PredSuccType = {}

        self._direct_supersuccs: _PredSuccType = {}
        self._direct_successors: _PredSuccType = {}
        self._direct_subsuccs: _PredSuccType = {}

        self._superpreds: _PredSuccType = {}
        self._predecessors: _PredSuccType = {}
        self._subpreds: _PredSuccType = {}

        self._supersuccs: _PredSuccType = {}
        self._successors: _PredSuccType = {}
        self._subsuccs: _PredSuccType = {}

        self._calc_direct_succs_preds()
        self._calc_predecessors()
        self._calc_successors()

    def components(self) -> Iterable[Component]:
        """Return the components of the model (nodes)."""
        return self._model.components

    def component(self, name: Reference) -> Component:
        """Return a component by name.

        Args:
            name: Name of the component to look up

        Returns:
            The component with the given name

        Raises:
            KeyError: If no component could be found
        """
        for component in self._model.components:
            if component.name == name:
                return component
        raise KeyError('Component {} not found'.format(name))

    def successors(self, component: Component) -> Set[Tuple[Component, int]]:
        """Return the successors of the given component.

        Args:
            component: The reference component

        Returns:
            The set of components for which a path exists from
            component's O_F to their F_INIT.

        Raises:
            KeyError: If the component is not in the model used to
                construct this object.
        """
        return self._successors[component]

    def predecessors(self, component: Component) -> Set[Tuple[Component, int]]:
        """Return the predecessors of the given component.

        Args:
            component: The reference component

        Returns:
            The set of components for which a path exists from
            their O_F to component's F_INIT.

        Raises:
            KeyError: If the component is not in the model used to
                construct this object.
        """
        return self._predecessors[component]

    def macros(self, component: Component) -> Set[Tuple[Component, int]]:
        """Return the macros of the given component.

        These are components that are both before the component's
        F_INIT and after its O_F.

        Args:
            component: The reference component

        Returns:
            The set of components that are both super-predecessor
            and super-successor of component.
        """
        return self._superpreds[component] & self._supersuccs[component]

    def micros(self, component: Component) -> Set[Tuple[Component, int]]:
        """Return the micros of the given component.

        These are components that are in between the component's
        O_I and its S.

        Args:
            component: The reference component

        Returns:
            The set of components that are both sub-successor
            and sub-predecessor of component.
        """
        return self._subsuccs[component] & self._subpreds[component]

    def _propagate(
            self, from_set: Set[Tuple[Component, int]],
            to_set: Set[Tuple[Component, int]], shared_dims: int
            ) -> None:
        """Propagates from_set into to_set.

        Note: Modifies to_set.

        Args:
            from_set: Set to propagate components from
            to_set: Set to propagate them into
            shared_dims: Maximum shared dimensions to carry
        """
        to_set.update({
            (cmp, min(sd, shared_dims))
            for cmp, sd in from_set})

    def _calc_predecessors(self) -> None:
        """Calculates predecessors of each component in the model.

        This function determines for each component in the model which
        other components can be reached by following incoming conduits.

        Preconditions:
            self._model set
            self._direct_* calculated

        Side effects:
            Fills self._superpreds, self._predecessors, self._subpreds
            with for each component in the model the corresponding set
            of components.
        """
        self._superpreds = {c: set() for c in self._model.components}
        self._predecessors = {c: set() for c in self._model.components}
        self._subpreds = {c: set() for c in self._model.components}

        num_remaining_preds = {
                c: (
                    len(self._direct_predecessors[c]) +
                    len(self._direct_superpreds[c]))
                for c in self._model.components}
        num_remaining_subpreds = {
                c: len(self._direct_subpreds[c])
                for c in self._model.components}

        todo = set(self._model.components)
        started: Set[Component] = set()
        doing: Set[Component] = set()
        finished: Set[Component] = set()
        done: Set[Component] = set()
        while todo or doing:
            started.clear()
            for component in todo:
                if num_remaining_preds[component] == 0:
                    for subsucc, shared_dims in self._direct_subsuccs[component]:
                        self._superpreds[subsucc].add((component, shared_dims))
                        self._propagate(
                                self._predecessors[component],
                                self._predecessors[subsucc], shared_dims)

                        self._propagate(
                                self._superpreds[component],
                                self._superpreds[subsucc], shared_dims)

                        num_remaining_preds[subsucc] -= 1
                    started.add(component)

            todo -= started
            doing |= started

            finished.clear()
            for component in doing:
                if num_remaining_subpreds[component] == 0:
                    for succ, shared_dims in self._direct_successors[component]:
                        self._predecessors[succ].add((component, shared_dims))
                        self._propagate(
                                self._subpreds[component],
                                self._predecessors[succ], shared_dims)

                        self._propagate(
                                self._predecessors[component],
                                self._predecessors[succ], shared_dims)

                        self._propagate(
                                self._superpreds[component],
                                self._superpreds[succ], shared_dims)

                        num_remaining_preds[succ] -= 1

                    for supersucc, shared_dims in self._direct_supersuccs[component]:
                        self._subpreds[supersucc].add((component, shared_dims))

                        self._propagate(
                                self._subpreds[component],
                                self._subpreds[supersucc], shared_dims)

                        self._propagate(
                                self._predecessors[component],
                                self._subpreds[supersucc], shared_dims)

                        num_remaining_subpreds[supersucc] -= 1

                    finished.add(component)

            doing -= finished
            done |= finished

            if not started and not finished:
                raise RuntimeError(
                        'Could not plan resource allocation for this model.'
                        ' Do you have a cycle of O_F -> F_INIT conduits?'
                        ' That does not work, because the models will all be'
                        ' waiting for each other to start.')

    def _calc_successors(self) -> None:
        """Calculates successors of each component in the model.

        This function determines for each component in the model which
        other components can be reached by following outgoing conduits.

        Preconditions:
            self._model set
            self._direct_* calculated

        Side effects:
            Fills self._supersuccs, self._successors, self._subsuccs
            with for each component in the model the corresponding set
            of components.
        """
        self._supersuccs = {c: set() for c in self._model.components}
        self._successors = {c: set() for c in self._model.components}
        self._subsuccs = {c: set() for c in self._model.components}

        num_remaining_succs = {
                c: (
                    len(self._direct_successors[c]) +
                    len(self._direct_supersuccs[c]))
                for c in self._model.components}

        num_remaining_subsuccs = {
                c: len(self._direct_subsuccs[c])
                for c in self._model.components}

        todo = set(self._model.components)
        started: Set[Component] = set()
        doing: Set[Component] = set()
        finished: Set[Component] = set()
        done: Set[Component] = set()
        while todo or doing:
            started.clear()
            for component in todo:
                if num_remaining_succs[component] == 0:
                    for subpred, shared_dims in self._direct_subpreds[component]:
                        self._supersuccs[subpred].add((component, shared_dims))
                        self._propagate(
                                self._successors[component],
                                self._successors[subpred], shared_dims)

                        self._propagate(
                                self._supersuccs[component],
                                self._supersuccs[subpred], shared_dims)
                        num_remaining_succs[subpred] -= 1

                    started.add(component)

            todo -= started
            doing |= started

            finished.clear()
            for component in doing:
                if num_remaining_subsuccs[component] == 0:
                    for pred, shared_dims in self._direct_predecessors[component]:
                        self._successors[pred].add((component, shared_dims))
                        self._propagate(
                                self._successors[component],
                                self._successors[pred], shared_dims)

                        self._propagate(
                                self._supersuccs[component],
                                self._supersuccs[pred], shared_dims)

                        self._propagate(
                                self._subsuccs[component],
                                self._successors[pred], shared_dims)
                        num_remaining_succs[pred] -= 1

                    for superpred, shared_dims in self._direct_superpreds[component]:
                        self._subsuccs[superpred].add((component, shared_dims))
                        self._propagate(
                                self._successors[component],
                                self._subsuccs[superpred], shared_dims)

                        self._propagate(
                                self._subsuccs[component],
                                self._subsuccs[superpred], shared_dims)
                        num_remaining_subsuccs[superpred] -= 1
                    finished.add(component)

            doing -= finished
            done |= finished

    def _calc_direct_succs_preds(self) -> None:
        """Calculates all successors and predecessors of components.

        Preconditions:
            self._model set

        Side effects:
            Sets self._direct_successors to a dictionary mapping each
            component in the model to the set of components that it
            has a * -> F_INIT conduit to.
        """
        components = {c.name: c for c in self._model.components}
        self._direct_supersuccs = {c: set() for c in self._model.components}
        self._direct_successors = {c: set() for c in self._model.components}
        self._direct_subsuccs = {c: set() for c in self._model.components}
        self._direct_superpreds = {c: set() for c in self._model.components}
        self._direct_predecessors = {c: set() for c in self._model.components}
        self._direct_subpreds = {c: set() for c in self._model.components}

        for conduit in self._model.conduits:
            sender = components[conduit.sending_component()]
            snd_op = None
            if sender.ports:
                if conduit.sending_port() in sender.ports.o_i:
                    snd_op = Operator.O_I
                elif conduit.sending_port() in sender.ports.o_f:
                    snd_op = Operator.O_F

            receiver = components[conduit.receiving_component()]
            recv_op = None
            if conduit.receiving_port() == 'muscle_settings_in':
                recv_op = Operator.F_INIT
            elif receiver.ports:
                if conduit.receiving_port() in receiver.ports.f_init:
                    recv_op = Operator.F_INIT
                elif conduit.receiving_port() in receiver.ports.s:
                    recv_op = Operator.S

            shared_dims = min(len(sender.multiplicity), len(receiver.multiplicity))

            if (snd_op, recv_op) == (Operator.O_I, Operator.F_INIT):
                self._direct_superpreds[receiver].add((sender, shared_dims))
                self._direct_subsuccs[sender].add((receiver, shared_dims))
            elif (snd_op, recv_op) == (Operator.O_F, Operator.F_INIT):
                self._direct_successors[sender].add((receiver, shared_dims))
                self._direct_predecessors[receiver].add((sender, shared_dims))
            elif (snd_op, recv_op) == (Operator.O_F, Operator.S):
                self._direct_subpreds[receiver].add((sender, shared_dims))
                self._direct_supersuccs[sender].add((receiver, shared_dims))


class ResourceAssignment:
    """Assigned resources for each process of an instance.

    Note that we use the classes from libmuscle.planner.resources to generically refer
    to collections of resources, either to describe the available hardware or to
    designate a subset of it that is occupied by a particular instance, or a subset that
    isn't currently occupied.

    This class has more detailed information, because it knows for each process (MPI
    rank) in the instance which subset of the overall resources for the instance it
    should be on, which we need to launch it in the right place.

    Attributes:
        by_rank: List of OnNodeResources objects containing assigned resources,
        indexed by rank.
    """
    def __init__(self, by_rank: List[OnNodeResources]) -> None:
        """Create a ResourceAssignment.

        Args:
            by_rank: List of OnNodeResources objects containing assigned resources,
            indexed by rank.
        """
        self.by_rank = by_rank

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ResourceAssignment):
            return NotImplemented

        return (
                len(self.by_rank) == len(other.by_rank) and
                all([
                    snr == onr
                    for snr, onr in zip(self.by_rank, other.by_rank)]))

    def __str__(self) -> str:
        # str(list()) uses repr() on the elements, we want str()
        str_rbr = ', '.join([str(nr) for nr in self.by_rank])
        return f'[{str_rbr}]'

    def __repr__(self) -> str:
        return f'ResourceAssignment({repr(self.by_rank)})'

    def as_resources(self) -> Resources:
        """Return a Resources representing the combined assigned resources."""
        result = Resources()
        for node_res in self.by_rank:
            result.merge_node(node_res)
        return result


class InsufficientResourcesAvailable(RuntimeError):
    pass


class Planner:
    """Allocates resources and keeps track of allocations."""
    def __init__(self, all_resources: Resources) -> None:
        """Create a Planner.

        Args:
            all_resources: An object describing the available resources
                    for the planner to use.
        """
        self._all_resources = all_resources
        self._allocations: Dict[Reference, Resources] = {}
        self._oversubscribed: Dict[Reference, Resources] = {}
        self._next_virtual_node = 1

    def allocate_all(
            self, configuration: Configuration, virtual: bool = False
            ) -> Dict[Reference, ResourceAssignment]:
        """Allocates resources for the given components.

        Allocation can occur either on a fixed set of available
        resources (virtual set to  False), or on an elastic set of
        virtual resources (virtual set to True). Use the former
        inside of a (fixed) cluster allocation and the latter to
        determine how many nodes are needed to run a simulation
        without oversubscribing.

        If virtual is True, additional nodes will be added as
        needed to obtain the resources needed to allocate all
        instances. Each additional node will have as many cores
        as a random existing one. The intended use is to pass a
        single node to __init__ when using this mode.

        Args:
            configuration: Configuration to allocate all components of
            virtual: Allocate on virtual resources or not, see above

        Returns:
            Assigned resources for each instance required by the model.
        """
        result: Dict[Reference, ResourceAssignment] = {}

        _logger.debug(f'Planning on resources {self._all_resources}')

        # Analyse model
        model = ModelGraph(configuration.model)
        requirements = configuration.resources
        implementations = configuration.implementations
        exclusive = {
                i for c in model.components() for i in c.instances()
                if (c.implementation and
                    not implementations[c.implementation].can_share_resources)}

        # Allocate
        unallocated_instances = [
                i for c in model.components() for i in c.instances()]
        leftover_instances: List[Reference] = []
        while unallocated_instances:
            leftover_instances.clear()

            to_allocate = self._sort_instances(
                    unallocated_instances, requirements)

            for instance in to_allocate:
                _logger.debug(f'Placing {instance}')
                component = model.component(instance.without_trailing_ints())
                conflicting_names = self._conflicting_names(
                        model, exclusive, component, instance)

                done = False
                while not done:
                    try:
                        result[instance] = self._assign_instance(
                                instance, component,
                                requirements[component.name],
                                conflicting_names, virtual)
                        done = True
                    except InsufficientResourcesAvailable:
                        if virtual:
                            self._expand_resources(
                                    component.name,
                                    requirements[component.name])
                        else:
                            leftover_instances.append(instance)
                            done = True

            if leftover_instances:
                _logger.warning(
                    'Planner ran out of resources, oversubscribing remaining'
                    ' instances.')
                self._oversubscribed.update(self._allocations)
                self._allocations.clear()

            unallocated_instances.clear()
            unallocated_instances.extend(leftover_instances)

        return result

    def _sort_instances(
            self, instances: List[Reference],
            requirements: Mapping[Reference, ResourceRequirements]
            ) -> List[Reference]:
        """Return to be allocated components in optimal order.

        This is a heuristic, it's not actually optimal but it should
        give decent results most of the time.

        Args:
            instances: The instances to sort
            requirements: The resource requirements for their
                components, indexed by name
        """
        cmp_names = map(Reference.without_trailing_ints, instances)
        reqs = map(lambda n: requirements[n], cmp_names)
        instances_reqs = list(zip(instances, reqs))
        threaded = [
                (i, r.threads) for i, r in instances_reqs
                if isinstance(r, ThreadedResReq)]
        sorted_threaded = sorted(threaded, key=lambda x: x[1], reverse=True)
        sorted_threaded_instances = [x[0] for x in sorted_threaded]

        mpi = [
                (i, r.mpi_processes) for i, r in instances_reqs
                if isinstance(r, MPICoresResReq)]
        sorted_mpi = sorted(mpi, key=lambda x: x[1], reverse=True)
        sorted_mpi_instances = [x[0] for x in sorted_mpi]

        return sorted_threaded_instances + sorted_mpi_instances

    def _conflicting_names(
            self, model: ModelGraph, exclusive: Set[Reference],
            component: Component, instance: Reference) -> Set[Reference]:
        """Find conflicting components.

        This returns the names of instances that cannot share resources
        with the given component, so that they can be avoided when
        assigning resources.

        Args:
            model: Model to search
            exclusive: List of instances that cannot share resources
            component: Component to find conflicts for
            instance: Instance (of component) to find conflicts for
        """
        def indices_match(
                instance1: Reference, instance2: Reference, dims: int) -> bool:
            idx1 = instance_indices(instance1)
            idx2 = instance_indices(instance2)
            return idx1[:dims] == idx2[:dims]

        def matching_instances(
                others: Set[Tuple[Component, int]]) -> Set[Reference]:
            return {
                    i for c, d in others for i in c.instances()
                    if indices_match(i, instance, d)}

        conflicting_instances = {
                i for c in model.components() for i in c.instances()}

        if instance in exclusive:
            return conflicting_instances

        conflicting_instances -= matching_instances(model.predecessors(component))
        conflicting_instances -= matching_instances(model.successors(component))

        if component not in exclusive:
            micros = matching_instances(model.micros(component))
            macros = matching_instances(model.macros(component))
            nonconflicting_mms = (micros | macros) - exclusive
            conflicting_instances -= nonconflicting_mms

        return conflicting_instances

    def _expand_resources(
            self, name: Reference, req: ResourceRequirements) -> None:
        """Adds an extra virtual node to the available resources."""
        taken = True
        while taken:
            new_node_name = 'node{:06d}'.format(self._next_virtual_node)
            taken = new_node_name in self._all_resources.nodes()
            self._next_virtual_node += 1

        new_node = copy(next(iter(self._all_resources)))
        new_node.node_name = new_node_name

        num_cores = len(new_node.cpu_cores)
        if isinstance(req, ThreadedResReq):
            if req.threads > num_cores:
                raise InsufficientResourcesAvailable(
                        f'Instance {name} requires {req.threads} threads,'
                        f' which is impossible with {num_cores} cores per'
                        ' node.')
        if isinstance(req, MPICoresResReq):
            if req.threads_per_mpi_process > num_cores:
                raise InsufficientResourcesAvailable(
                        f'Instance {name} requires'
                        f' {req.threads_per_mpi_process} threads per process,'
                        f' which is impossible with {num_cores} cores per'
                        ' node.')

        self._all_resources.add_node(new_node)

    def _assign_instance(
            self, instance: Reference, component: Component,
            requirements: ResourceRequirements,
            simultaneous_instances: Set[Reference], virtual: bool
            ) -> ResourceAssignment:
        """Allocates resources for the given instance.

        If we are on real resources, and the instance requires more
        threads than our nodes have cores, then we'll just give it all
        cores on a node and hope for the best. If we are on virtual
        resources, this will raise InsufficientResourcesAvailable.

        Args:
            instance: The instance to assign resources to
            component: The component it is an instance of
            requirements: Its resource requirements
            simultaneous_instances: Instances which may execute
                    simultaneously and whose resources we therefore
                    cannot overlap with
            virtual: Whether we are on virtual resources

        Returns:
            The resources assigned to each process in the instance
        """
        assignment = ResourceAssignment([])
        free_resources = copy(self._all_resources)

        for other in self._allocations:
            if other in simultaneous_instances:
                free_resources -= self._allocations[other]

        _logger.debug(f'Free resources: {free_resources}')
        try:
            if isinstance(requirements, ThreadedResReq):
                assignment.by_rank.append(self._assign_thread_block(
                        free_resources, requirements.threads))

            elif isinstance(requirements, MPICoresResReq):
                if requirements.threads_per_mpi_process != 1:
                    raise RuntimeError(
                            'Multiple threads per MPI process is not supported'
                            ' yet. Please make an issue on GitHub.')
                for proc in range(requirements.mpi_processes):
                    block = self._assign_thread_block(
                                free_resources, requirements.threads_per_mpi_process)
                    assignment.by_rank.append(block)
                    free_resources -= Resources([block])

            elif isinstance(requirements, MPINodesResReq):
                raise RuntimeError(
                        'Node-based MPI resource requirements are not'
                        ' supported yet. Please make an issue on GitHub.')

        except InsufficientResourcesAvailable:
            if not self._allocations and not virtual:
                # There are no other allocations and it's still not
                # enough. Just give it all and hope for the best.
                assignment = self._oversubscribe_instance(instance, requirements)
            else:
                raise

        self._allocations[instance] = assignment.as_resources()
        return assignment

    def _assign_thread_block(
            self, free_resources: Resources, num_threads: int) -> OnNodeResources:
        """Assign resources for a group of threads.

        This chooses a set of <num_threads> cores on the same node. It returns the
        assigned resources; it doesn't update self._allocations or free_resources.

        Args:
            num_threads: Number of threads to allocate for
            free_resources: Available resources to allocate from

        Returns:
            The assigned resources
        """
        for node in free_resources:
            if len(node.cpu_cores) >= num_threads:
                available_cores = node.cpu_cores
                _logger.debug(f'available cores: {available_cores}')
                to_reserve = available_cores.get_first_cores(num_threads)
                _logger.debug(f'assigned {to_reserve}')
                return OnNodeResources(node.node_name, to_reserve)
        raise InsufficientResourcesAvailable()

    def _oversubscribe_instance(
            self, instance: Reference, requirements: ResourceRequirements
            ) -> ResourceAssignment:
        """Oversubscribe an instance.

        This is called when all resources are available and we still cannot fit an
        instance, i.e. that single instance requires more resources than we have
        available in total. In that case, we're just going to map it onto the resources
        we have and hope for the best, which is what this function does.

        There's a lot of repetition between this and the code above. There's probably a
        cleaner way to do this, but it'll do for now. Eventually we'll have an optimiser
        and all this goes away anyway.

        Args:
            instance: The instance we're oversubscribing
            requirements: The required resources

        Returns:
            An oversubscribed resource assignment
        """
        _logger.warning(
                f'Instance {instance} requires more resources than are available in'
                ' total. Oversubscribing this instance.')

        res_by_rank: List[OnNodeResources] = list()

        if isinstance(requirements, ThreadedResReq):
            res_by_rank.append(copy(next(iter(self._all_resources))))

        elif isinstance(requirements, MPICoresResReq):
            if requirements.threads_per_mpi_process != 1:
                raise RuntimeError(
                        'Multiple threads per MPI process is not supported yet. Please'
                        ' make an issue on GitHub.')

            free_resources = copy(self._all_resources)
            for proc in range(requirements.mpi_processes):
                if free_resources.total_cores() < requirements.threads_per_mpi_process:
                    free_resources = copy(self._all_resources)

                block = self._assign_thread_block(
                            free_resources, requirements.threads_per_mpi_process)

                res_by_rank.append(block)
                free_resources -= Resources([block])

        return ResourceAssignment(res_by_rank)



from libmuscle.profiling import (
        ProfileEvent, ProfileEventType, ProfileTimestamp)

import time


def test_recording_events(mocked_profiler) -> None:
    profiler, _ = mocked_profiler

    t1 = ProfileTimestamp()
    t2 = ProfileTimestamp()
    e = ProfileEvent(ProfileEventType.REGISTER, t1, t2)

    profiler.record_event(e)

    assert e.start_time == t1
    assert e.stop_time == t2


def test_auto_stop_time(mocked_profiler) -> None:
    profiler, _ = mocked_profiler

    t1 = ProfileTimestamp()
    e = ProfileEvent(ProfileEventType.SEND, t1)

    profiler.record_event(e)

    assert e.start_time == t1
    assert e.stop_time is not None
    assert e.start_time.nanoseconds < e.stop_time.nanoseconds


def test_send_to_manager(profiler_comm_int_10ms, mocked_profiler) -> None:
    profiler, mock_mmp_client = mocked_profiler

    e1 = ProfileEvent(ProfileEventType.RECEIVE, ProfileTimestamp())
    profiler.record_event(e1)

    time.sleep(0.1)
    assert mock_mmp_client.sent_events == [e1]

    mock_mmp_client.sent_events = None
    e2 = ProfileEvent(ProfileEventType.RECEIVE, ProfileTimestamp())
    profiler.record_event(e2)

    time.sleep(0.1)
    assert mock_mmp_client.sent_events == [e2]

    mock_mmp_client.sent_events = None
    profiler.set_level('none')
    e3 = ProfileEvent(ProfileEventType.RECEIVE, ProfileTimestamp())
    profiler.record_event(e3)

    time.sleep(0.1)
    assert mock_mmp_client.sent_events is None



import sys
from collections import OrderedDict
from pathlib import Path
from typing import Sequence

import click
import ymmsl
from ymmsl import PartialConfiguration


from libmuscle.planner.planner import Planner, InsufficientResourcesAvailable
from libmuscle.planner.resources import Core, CoreSet, OnNodeResources, Resources
from libmuscle.snapshot_manager import SnapshotManager
from muscle3.profiling import (
        plot_instances, plot_resources, plot_timeline, show_plots)


_RESOURCES_INCOMPLETE_MODEL = """
A model, implementations and resources must be given to be able to calculate
resources for a model. Please ensure that you have specified all of them in
your yMMSL file(s).
"""


@click.group()
def muscle3() -> None:
    """MUSCLE3 command line interface

    This command provides various functions for running coupled simulations
    using MUSCLE3.

    Use muscle3 <command> --help for help with individual commands.
    """
    pass


@muscle3.command(short_help='Visualise profiling information')
@click.argument(
        'performance_file', nargs=1, type=click.Path(
            exists=True, file_okay=True, dir_okay=False, readable=True,
            allow_dash=True, resolve_path=True))
@click.option('-i', '--instances', is_flag=True, help='Show per-instance statistics')
@click.option('-r', '--resources', is_flag=True, help='Show per-resource statistics')
@click.option('-t', '--timeline', is_flag=True, help='Show a timeline of events')
def profile(
        performance_file: str, instances: bool, resources: bool, timeline: bool
        ) -> None:
    """Visualise profiling information.

    Takes data from a performance.sqlite file written by the manager,
    and produces one or more plots on the screen.

    Profiling functionality is still somewhat experimental, so this
    command may be changed and extended in the coming releases. Please
    let us know if there are any changes or additional features you
    would like to have by making an issue on GitHub, or through your
    usual support channel.

    Result:
        With -i, shows a chart depicting for each instance how much time
        it spend running, waiting and communicating.

        With -r, shows a chart depicting for each resource (CPU core)
        which instances ran on it for how long.

        With -t, shows a timeline of the different instances, depicting
        when they were sending, waiting, receiving and running.
    """
    if instances:
        plot_instances(Path(performance_file))

    if resources:
        plot_resources(Path(performance_file))

    if timeline:
        plot_timeline(Path(performance_file))
        click.echo(
                'Hint: To inspect the timeline in more detail, use the'
                ' pan and zoom buttons at the bottom of the window.')

    if instances or resources or timeline:
        show_plots()
    else:
        click.echo(
                'Please specify -i, -r or -t to choose which data'
                ' you would like to see.')
        sys.exit(1)

    sys.exit(0)


@muscle3.command(short_help='Calculate resources needed for a simulation')
@click.argument(
        'ymmsl_files', nargs=-1, required=True, type=click.Path(
            exists=True, file_okay=True, dir_okay=False, readable=True,
            allow_dash=True, resolve_path=True))
@click.option(
        '-c', '--cores-per-node', nargs=1, type=int, required=True,
        help='Set number of cores per cluster node.')
@click.option(
        '-v', '--verbose', is_flag=True, help='Show instance allocations.')
def resources(
        ymmsl_files: Sequence[str],
        cores_per_node: int, verbose: bool
        ) -> None:
    """Calculate the number of nodes needed to run the simulation.

    In order to run a MUSCLE3 simulation on a cluster, a batch job has
    to be submitted specifying the resources needed. This command
    calculates the number of nodes needed to run a given simulation
    without oversubscribing, given the number of cores available in
    each node.

    If multiple yMMSL files are given, then they will be combined left
    to right, i.e. if there are conflicting declarations, the one from
    the file last given is used.

    Result:

      With the -v option, the allocation of each instance will be printed
      to stdout. Without it, a single number will be printed, the number
      of nodes needed to run the simulation.

    Examples:

      muscle3 resources --cores-per-node 24 simulation.ymmsl

      muscle3 resources -c 16 model.ymmsl resources.ymmsl

      num_nodes=$(muscle3 resources -c 16 simulation.ymmsl)
      sbatch -N ${num_nodes} simulation.sh

    """
    partial_config = _load_ymmsl_files(ymmsl_files)
    try:
        config = partial_config.as_configuration()
    except ValueError:
        click.echo(_RESOURCES_INCOMPLETE_MODEL, err=True)
        sys.exit(1)

    resources = Resources([
        OnNodeResources(
            'node000001', CoreSet([Core(i, {i}) for i in range(cores_per_node)]))])

    planner = Planner(resources)
    try:
        allocations = planner.allocate_all(config, True)
    except InsufficientResourcesAvailable as e:
        click.echo(e)
        sys.exit(1)

    num_nodes = len(tuple(resources.nodes()))

    if verbose:
        click.echo()
        if num_nodes == 1:
            click.echo('A total of 1 node will be needed, as follows:')
        else:
            click.echo(
                    f'A total of {num_nodes} nodes will be needed,'
                    ' as follows:')

        click.echo()
        for instance in sorted(allocations):
            click.echo(f'{instance}: {str(allocations[instance])}')
    else:
        click.echo(f'{num_nodes}', nl=False)

    sys.exit(0)


@muscle3.command(short_help='Display details of a stored snapshot')
@click.argument(
        'snapshot_files', nargs=-1, required=True, type=click.Path(
            exists=True, file_okay=True, dir_okay=False, readable=True,
            allow_dash=True, resolve_path=True, path_type=Path))
@click.option(
        '-d', '--data', is_flag=True,
        help='Display stored data. Note this may result in a lot of output!')
@click.option(
        '-v', '--verbose', is_flag=True, help='Display more metadata.')
def snapshot(
        snapshot_files: Sequence[Path], data: bool, verbose: bool) -> None:
    """Display information about stored snapshots.

    Per provided snapshot, display metadata. Stored data can also be output by
    supplying the '-d' or '--data' flags. Note that this may result in a lot of
    data displayed.
    """
    for file in snapshot_files:
        snapshot = SnapshotManager.load_snapshot_from_file(file)
        click.echo(f'Snapshot at {file}:')
        typ = 'Final' if snapshot.is_final_snapshot else 'Intermediate'
        properties = OrderedDict([
            ('Snapshot type', typ),
            ('Snapshot timestamp',
             snapshot.message.timestamp if snapshot.message else float('-inf')),
            ('Snapshot wallclock time', snapshot.wallclock_time),
            ('Snapshot triggers', snapshot.triggers),
        ])
        if verbose:
            properties.update([
                ('Internal: Port message counts', snapshot.port_message_counts),
            ])
        for prop_name, prop_value in properties.items():
            click.secho(f'{prop_name}: ', nl=False, bold=True)
            click.echo(prop_value)
        if data:
            click.secho('Snapshot data:', bold=True)
            if snapshot.message is not None:
                click.echo(snapshot.message.data)
            else:
                click.secho("No data available", italic=True)
        click.echo()


def _load_ymmsl_files(ymmsl_files: Sequence[str]) -> PartialConfiguration:
    """Loads and merges yMMSL files."""
    configuration = PartialConfiguration()
    for path in ymmsl_files:
        with open(path, 'r') as f:
            configuration.update(ymmsl.load(f))
    return configuration


if __name__ == '__main__':
    muscle3()



from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List


class AgentCommand:
    pass


@dataclass
class StartCommand(AgentCommand):
    name: str
    work_dir: Path
    args: List[str]
    env: Dict[str, str]
    stdout: Path
    stderr: Path


class CancelAllCommand(AgentCommand):
    pass


class ShutdownCommand(AgentCommand):
    pass



import errno
import logging
from typing import Any, Dict, cast, List, Optional

import msgpack

from libmuscle.mcp.protocol import AgentCommandType, RequestType, ResponseType
from libmuscle.mcp.tcp_transport_server import TcpTransportServer
from libmuscle.mcp.transport_server import RequestHandler
from libmuscle.native_instantiator.agent.agent_commands import (
        AgentCommand, CancelAllCommand, ShutdownCommand, StartCommand)
from libmuscle.native_instantiator.iagent_manager import IAgentManager
from libmuscle.planner.resources import Core, CoreSet, OnNodeResources
from libmuscle.post_office import PostOffice

from ymmsl import Reference


_logger = logging.getLogger(__name__)


class MAPRequestHandler(RequestHandler):
    """Handles Agent requests."""
    def __init__(self, agent_manager: IAgentManager, post_office: PostOffice) -> None:
        """Create a MAPRequestHandler.

        Args:
            agent_manager: The AgentManager to forward reports to
            post_office: The PostOffice to get commands from
        """
        self._agent_manager = agent_manager
        self._post_office = post_office

    def handle_request(self, request: bytes) -> bytes:
        """Handles an agent request.

        Args:
            request: The encoded request

        Returns:
            response: An encoded response
        """
        req_list = msgpack.unpackb(request, raw=False)
        req_type = req_list[0]
        req_args = req_list[1:]
        if req_type == RequestType.REPORT_RESOURCES.value:
            response = self._report_resources(*req_args)
        elif req_type == RequestType.GET_COMMAND.value:
            response = self._get_command(*req_args)
        elif req_type == RequestType.REPORT_RESULT.value:
            response = self._report_result(*req_args)

        return cast(bytes, msgpack.packb(response, use_bin_type=True))

    def _report_resources(
            self, node_name: str, data: Dict[str, Any]) -> Any:
        """Handle a report resources request.

        This is used by the agent to report available resources on its node when
        it starts up.

        Args:
            node_name: Name (hostname) of the node
            data: Resource dictionary, containing a single key 'cpu' which maps to a
                list of cores, where each core is a list of ints, starting with the core
                id at index [0] followed by the hwthread ids of all hwthreads in this
                core.
        """
        cores = CoreSet((Core(ids[0], set(ids[1:])) for ids in data['cpu']))
        node_resources = OnNodeResources(node_name, cores)
        self._agent_manager.report_resources(node_resources)
        return [ResponseType.SUCCESS.value]

    def _get_command(self, node_name: str) -> Any:
        """Handle a get command request.

        This is used by the agent to ask if there's anything we would like it to do.
        Command sounds a bit brusque, but we already have the agent sending requests
        to this handler, so I needed a different word to distinguish them. Requests
        are sent by the agent to the manager (because it's the client in an RPC setup),
        commands are returned by the manager to the agent (because it tells it what to
        do).

        Args:
            node_name: Hostname (name) of the agent's node
        """
        node_ref = Reference('_' + node_name.replace('-', '_'))
        next_request: Optional[bytes] = None
        if self._post_office.have_message(node_ref):
            next_request = self._post_office.get_message(node_ref)

        if next_request is not None:
            return [ResponseType.SUCCESS.value, next_request]

        return [ResponseType.PENDING.value]

    def _report_result(self, instances: List[List[Any]]) -> Any:
        """Handle a report result rquest.

        This is sent by the agent if an instance it launched exited.

        Args:
            instances: List of instance descriptions, comprising an id str and exit
                    code int. Really a List[Tuple[str, int]] but msgpack doesn't know
                    about tuples.
        """
        self._agent_manager.report_result(list(map(tuple, instances)))
        return [ResponseType.SUCCESS.value]


class MAPServer:
    """The MUSCLE Agent Protocol server.

    This class accepts connections from the agents and services them using a
    MAPRequestHandler.
    """
    def __init__(self, agent_manager: IAgentManager) -> None:
        """Create a MAPServer.

        This starts a TCP Transport server and connects it to a MAPRequestHandler,
        which uses the given agent manager to service the requests. By default, we
        listen on port 9009, unless it's not available in which case we use a random
        other one.

        Args:
            agent_manager: AgentManager to forward requests to
        """
        self._post_office = PostOffice()
        self._handler = MAPRequestHandler(agent_manager, self._post_office)
        try:
            self._server = TcpTransportServer(self._handler, 9009)
        except OSError as e:
            if e.errno != errno.EADDRINUSE:
                raise
            self._server = TcpTransportServer(self._handler)

    def get_location(self) -> str:
        """Return this server's network location.

        This is a string of the form tcp:<hostname>:<port>.
        """
        return self._server.get_location()

    def stop(self) -> None:
        """Stop the server.

        This makes the server stop serving requests, and shuts down its
        background threads.
        """
        self._server.close()

    def deposit_command(self, node_name: str, command: AgentCommand) -> None:
        """Deposit a command for the given agent.

        This takes the given command and queues it for the given agent to pick up next
        time it asks us for one.

        Args:
            node_name: Name of the node whose agent should execute the command
            command: The command to send
        """
        agent = Reference('_' + node_name.replace('-', '_'))

        if isinstance(command, StartCommand):
            command_obj = [
                    AgentCommandType.START.value, command.name, str(command.work_dir),
                    command.args, command.env, str(command.stdout), str(command.stderr)
                    ]
        elif isinstance(command, CancelAllCommand):
            command_obj = [AgentCommandType.CANCEL_ALL.value]
        elif isinstance(command, ShutdownCommand):
            command_obj = [AgentCommandType.SHUTDOWN.value]

        encoded_command = cast(bytes, msgpack.packb(command_obj, use_bin_type=True))

        self._post_office.deposit(agent, encoded_command)



from pathlib import Path
from typing import Optional

from ymmsl import Reference


class RunDir:
    """Manages a run directory containing files for a simulation run.

    The directory is laid out as follows:

    .. code-block::

        <run_dir>/
            input/
            1_<name>.ymmsl
            2_<name>.ymmsl
            muscle_manager.log
            muscle_stats.sqlite
            instances/
                <instance_name[i]>/
                    run_script.sh
                    <instance_name[i]>.out
                    <instance_name[i]>.err
                    work_dir/
                    snapshots/
            snapshots/

    """
    def __init__(self, run_dir: Path) -> None:
        """Create a RunDir managing the given directory.

        This creates the run dir if it does not exist.
        """
        self.path = run_dir

        if not self.path.exists():
            self.path.mkdir()

        instances_dir = self.path / 'instances'
        if not instances_dir.exists():
            instances_dir.mkdir()

    def add_instance_dir(self, name: Reference) -> Path:
        """Add a directory for this instance to the run dir.

        Args:
            name: Name of the instance

        Returns:
            The path to the new, empty directory
        """
        idir = self.path / 'instances' / str(name)
        if idir.exists():
            raise ValueError('Instance already has a directory')
        idir.mkdir()
        return idir

    def instance_dir(self, name: Reference) -> Path:
        """Return the instance dir of the given instance.

        The directory may or may not exist, call add_instance_dir() to
        make it.
        """
        return self.path / 'instances' / str(name)

    def snapshot_dir(self, name: Optional[Reference] = None) -> Path:
        """Return the snapshot directory for the workflow or for an instance.

        Args:
            name: Name of the instance. May be None to get the workflow snapshot
                directory.

        Returns:
            The path to the snapshot directory
        """
        if name is None:
            path = self.path / 'snapshots'
        else:
            path = self.instance_dir(name) / 'snapshots'
        path.mkdir(parents=True, exist_ok=True)
        return path



import logging
from threading import Lock
from typing import Dict, List, Optional, Tuple


_logger = logging.getLogger(__name__)


class DeadlockDetector:
    """The DeadlockDetector attempts to detect when multiple instances are stuck waiting
    for each other.

    This class is responsible for handling WAITING_FOR_RECEIVE, IS_DEADLOCKED and
    WAITING_FOR_RECEIVE_DONE MMP messages, which are submitted by the MMPServer.

    When a deadlock is detected, the cycle of instances that is waiting for each other
    is logged with FATAL severity.
    """

    def __init__(self) -> None:
        """Construct a new DeadlockDetector."""
        self._mutex = Lock()
        """Mutex that should be locked before accessing instance variables."""
        self._waiting_instances: Dict[str, str] = {}
        """Maps instance IDs to the peer instance IDs they are waiting for."""
        self._waiting_instance_ports: Dict[str, Tuple[str, Optional[int]]] = {}
        """Maps instance IDs to the port/slot they are waiting for.."""
        self._detected_deadlocks: List[List[str]] = []
        """List of deadlocked instance cycles. Set by _handle_potential_deadlock."""

    def waiting_for_receive(
            self, instance_id: str, peer_instance_id: str,
            port_name: str, slot: Optional[int]
            ) -> None:
        """Process a WAITING_FOR_RECEIVE message from an instance.

        This method can be called from any thread.

        Args:
            instance_id: ID of instance that is waiting to receive a message.
            peer_instance_id: ID of the peer that the instance is waiting for.
            port_name: Name of the input port.
            slot: Optional slot number of the input port.
        """
        with self._mutex:
            # Sanity checks, triggering this is a bug in the instance or the manager
            assert instance_id not in self._waiting_instances

            # Register that the instance is waiting
            self._waiting_instances[instance_id] = peer_instance_id
            self._waiting_instance_ports[instance_id] = (port_name, slot)
            self._check_for_deadlock(instance_id)

    def waiting_for_receive_done(
            self, instance_id: str, peer_instance_id: str,
            port_name: str, slot: Optional[int]
            ) -> None:
        """Process a WAITING_FOR_RECEIVE_DONE message from an instance.

        This method can be called from any thread.

        Args:
            instance_id: ID of instance that is waiting to receive a message.
            peer_instance_id: ID of the peer that the instance is waiting for.
            port_name: Name of the input port.
            slot: Optional slot number of the input port.
        """
        with self._mutex:
            # Sanity checks, triggering these is a bug in the instance or the manager
            assert instance_id in self._waiting_instances
            assert self._waiting_instances[instance_id] == peer_instance_id
            assert self._waiting_instance_ports[instance_id] == (port_name, slot)

            # We're not waiting anymore
            del self._waiting_instances[instance_id]
            del self._waiting_instance_ports[instance_id]

            # Check if we were part of a deadlock
            for i, instance_list in enumerate(self._detected_deadlocks):
                if instance_id in instance_list:
                    del self._detected_deadlocks[i]
                    break

    def is_deadlocked(self, instance_id: str) -> bool:
        """Check if the provided instance is part of a detected deadlock.

        This method can be called from any thread.
        """
        with self._mutex:
            for deadlock_instances in self._detected_deadlocks:
                if instance_id in deadlock_instances:
                    _logger.fatal(
                            "Deadlock detected, simulation is aborting!\n%s",
                            self._format_deadlock(deadlock_instances))
                    return True
        return False

    def _check_for_deadlock(self, instance_id: str) -> None:
        """Check if there is a cycle of waiting instances that involves this instance.

        Make sure to lock self._mutex before calling this.
        """
        deadlock_instances = [instance_id]
        cur_instance = instance_id
        while cur_instance in self._waiting_instances:
            cur_instance = self._waiting_instances[cur_instance]
            if cur_instance == instance_id:
                self._handle_potential_deadlock(deadlock_instances)
                return
            deadlock_instances.append(cur_instance)
        _logger.debug("No deadlock detected")

    def _handle_potential_deadlock(self, deadlock_instances: List[str]) -> None:
        """Handle a potential deadlock.

        Make sure to lock self._mutex before calling this.

        Args:
            deadlock_instances: list of instances waiting for eachother
        """
        self._detected_deadlocks.append(deadlock_instances)

    def _format_deadlock(self, deadlock_instances: List[str]) -> str:
        """Create and return formatted deadlock debug info.

        Args:
            deadlock_instances: list of instances waiting for eachother
        """
        num_instances = str(len(deadlock_instances))
        lines = [f"The following {num_instances} instances are deadlocked:"]
        for i, instance in enumerate(deadlock_instances):
            num = str(i+1).rjust(len(num_instances))
            peer_instance = self._waiting_instances[instance]
            port, slot = self._waiting_instance_ports[instance]
            slot_txt = "" if slot is None else f"[{slot}]"
            lines.append(
                f"{num}. Instance '{instance}' is waiting for instance"
                f" '{peer_instance}' in a receive on port '{port}{slot_txt}'."
            )
        return "\n".join(lines)



from enum import Enum
import logging

from libmuscle.timestamp import Timestamp


class LogLevel(Enum):
    """Log levels for MUSCLE3.

    These match the levels in the MUSCLE Manager Protocol, and should
    be kept identical to those. They also match the Python logging log
    levels.
    """
    DISABLE = 100
    CRITICAL = 50
    ERROR = 40
    WARNING = 30
    INFO = 20
    DEBUG = 10
    LOCAL = 0

    @staticmethod
    def from_python_level(level: int) -> 'LogLevel':
        """Creates a LogLevel from a Python log level.

        Args:
            level: A standard Python log level, such as logging.WARNING.

        Returns:
            The same level, but as a LogLevel.
        """
        if level > 40:
            return LogLevel.CRITICAL
        elif level > 30:
            return LogLevel.ERROR
        elif level > 20:
            return LogLevel.WARNING
        elif level > 10:
            return LogLevel.INFO
        return LogLevel.DEBUG

    def as_python_level(self) -> int:
        """Convert the LogLevel to the corresponding Python level."""
        to_python_level = {
                LogLevel.CRITICAL: logging.CRITICAL,
                LogLevel.ERROR: logging.ERROR,
                LogLevel.WARNING: logging.WARNING,
                LogLevel.INFO: logging.INFO,
                LogLevel.DEBUG: logging.DEBUG,
                LogLevel.LOCAL: logging.DEBUG}
        return to_python_level[self]


class LogMessage:
    """A log message as used by MUSCLE3.

    Args:
        instance_id: The identifier of the instance that generated \
                this message.
        timestamp: When the message was generated (real-world, not \
                simulation).
        level: Log level of the message.
        text: Content of the message.

    Attributes:
        instance_id: The identifier of the instance that generated \
                this message.
        timestamp: When the message was generated (real-world, not \
                simulation).
        level: Log level of the message.
        text: Content of the message.
    """
    def __init__(
            self,
            instance_id: str,
            timestamp: Timestamp,
            level: LogLevel,
            text: str
            ) -> None:

        self.instance_id = instance_id
        self.timestamp = timestamp
        self.level = level
        self.text = text



from contextlib import nullcontext as does_not_raise
import logging
from unittest.mock import MagicMock, patch

import pytest

from libmuscle.instance import Instance, InstanceFlags
from libmuscle.mpp_message import ClosePort
from ymmsl import Checkpoints, Operator, Reference as Ref, Settings


@pytest.fixture
def logger():
    with patch('libmuscle.instance._logger') as logger:
        yield logger


@pytest.fixture(autouse=True)
def MMPClient():
    with patch('libmuscle.instance.MMPClient') as MMPClient:
        mmp_client = MMPClient.return_value
        mmp_client.request_peers.return_value = (MagicMock(), MagicMock(), MagicMock())

        checkpoints = MagicMock()
        checkpoints.__bool__.return_value = False
        mmp_client.get_checkpoint_info.return_value = [
                MagicMock(), checkpoints, MagicMock(), MagicMock()]
        yield MMPClient


@pytest.fixture
def mmp_client(MMPClient):
    return MMPClient.return_value


@pytest.fixture
def api_guard():
    with patch('libmuscle.instance.APIGuard') as APIGuard:
        yield APIGuard.return_value


@pytest.fixture
def profiler():
    with patch('libmuscle.instance.Profiler') as Profiler:
        yield Profiler.return_value


@pytest.fixture
def port_manager():
    with patch('libmuscle.instance.PortManager') as PortManager:
        port_manager = PortManager.return_value
        port_manager.settings_in_connected.return_value = False
        yield port_manager


@pytest.fixture
def communicator():
    with patch('libmuscle.instance.Communicator') as Communicator:
        yield Communicator.return_value


@pytest.fixture
def settings_manager():
    with patch('libmuscle.instance.SettingsManager') as SettingsManager:
        settings_manager = SettingsManager.return_value
        # Emulate no settings available
        settings_manager.get_setting.side_effect = KeyError()
        yield settings_manager


@pytest.fixture(autouse=True)
def snapshot_manager():
    with patch('libmuscle.instance.SnapshotManager') as SnapshotManager:
        yield SnapshotManager.return_value


@pytest.fixture(autouse=True)
def no_resume_snapshot_manager(snapshot_manager):
    snapshot_manager.resuming_from_intermediate.return_value = False
    snapshot_manager.resuming_from_final.return_value = False
    snapshot_manager.resume_overlay = None
    return snapshot_manager


@pytest.fixture(autouse=True)
def trigger_manager():
    with patch('libmuscle.instance.TriggerManager') as TriggerManager:
        yield TriggerManager.return_value


@pytest.fixture
def sys_argv():
    with patch('libmuscle.instance.sys.argv', []) as sys_argv:
        yield sys_argv


@pytest.fixture
def manager_location_argv(sys_argv):
    sys_argv.extend(['--test', '--muscle-manager=tcp:localhost:9001', 'bla'])
    yield None


@pytest.fixture
def instance_argv(sys_argv):
    sys_argv.extend(['distraction', '--muscle-instance=component'])
    yield None


@pytest.fixture
def os_environ():
    with patch('libmuscle.instance.os.environ', {}) as os_environ:
        yield os_environ


@pytest.fixture
def manager_location_envvar(os_environ):
    os_environ['MUSCLE_MANAGER'] = 'tcp:localhost:9002'
    yield None


@pytest.fixture
def instance_envvar(os_environ):
    os_environ['MUSCLE_INSTANCE'] = 'component[13]'
    yield None


@pytest.fixture
def instance(
        logger, MMPClient, mmp_client, api_guard, profiler, connected_port_manager,
        communicator, settings_manager, no_resume_snapshot_manager, trigger_manager,
        manager_location_argv, instance_argv, declared_ports):

    return Instance(declared_ports)


@pytest.fixture
def instance_dont_apply_overlay(
        logger, MMPClient, mmp_client, api_guard, profiler, connected_port_manager,
        communicator, settings_manager, no_resume_snapshot_manager, trigger_manager,
        manager_location_argv, instance_argv, declared_ports):

    return Instance(declared_ports, InstanceFlags.DONT_APPLY_OVERLAY)


def test_create_instance_manager_location_default(
        instance_argv, MMPClient, declared_ports):

    instance = Instance(declared_ports)
    instance.error_shutdown("")  # ensure all threads and resources are cleaned up

    MMPClient.assert_called_once_with(Ref('component'), 'tcp:localhost:9000')


def test_create_instance_manager_location_argv(
        manager_location_argv, instance_argv, MMPClient, declared_ports):

    instance = Instance(declared_ports)
    instance.error_shutdown("")  # ensure all threads and resources are cleaned up

    MMPClient.assert_called_once_with(Ref('component'), 'tcp:localhost:9001')


def test_create_instance_manager_location_envvar(
        manager_location_envvar, instance_envvar, MMPClient, declared_ports):

    instance = Instance(declared_ports)
    instance.error_shutdown("")  # ensure all threads and resources are cleaned up

    MMPClient.assert_called_once_with(Ref('component[13]'), 'tcp:localhost:9002')


def test_create_instance_registration(
        manager_location_argv, instance_argv, mmp_client, communicator, port_manager,
        profiler, declared_ports):

    instance = Instance(declared_ports)
    instance.error_shutdown("")  # ensure all threads and resources are cleaned up

    locations = communicator.get_locations.return_value

    mmp_client.register_instance.assert_called_once()
    assert mmp_client.register_instance.call_args[0][0] == locations
    port_desc = mmp_client.register_instance.call_args[0][1]
    assert port_desc[0].name == 'in'
    assert port_desc[0].operator == Operator.F_INIT
    assert port_desc[1].name == 'not_connected'
    assert port_desc[1].operator == Operator.F_INIT
    assert port_desc[2].name == 'out_v'
    assert port_desc[2].operator == Operator.O_I
    assert port_desc[3].name == 'out_r'
    assert port_desc[3].operator == Operator.O_I
    assert port_desc[4].name == 'in_v'
    assert port_desc[4].operator == Operator.S
    assert port_desc[5].name == 'in_r'
    assert port_desc[5].operator == Operator.S
    assert port_desc[6].name == 'not_connected_v'
    assert port_desc[6].operator == Operator.S
    assert port_desc[7].name == 'out'
    assert port_desc[7].operator == Operator.O_F


def test_create_instance_profiling(
        manager_location_argv, instance_argv, profiler, declared_ports):

    instance = Instance(declared_ports)

    assert len(profiler.record_event.mock_calls) == 2
    instance.error_shutdown("Ensure all threads and resources are cleaned up")


def test_create_instance_connecting(
        manager_location_argv, instance_argv, mmp_client, port_manager, communicator,
        settings_manager, declared_ports):

    conduits = [MagicMock(), MagicMock()]
    peer_dims = {'component': [], 'other': []}
    peer_locations = {
            'component': ['tcp:localhost:9003'],
            'other': ['tcp:localhost:9004']}
    mmp_client.request_peers.return_value = (conduits, peer_dims, peer_locations)

    settings = MagicMock()
    mmp_client.get_settings.return_value = settings

    instance = Instance(declared_ports)

    port_manager.connect_ports.assert_called_once()
    peer_info = port_manager.connect_ports.call_args[0][0]
    assert peer_info._PeerInfo__kernel == Ref('component')
    assert peer_info._PeerInfo__index == []
    assert peer_info._PeerInfo__peer_dims == peer_dims
    assert peer_info._PeerInfo__peer_locations == peer_locations

    communicator.set_peer_info.assert_called_once()
    assert communicator.set_peer_info.call_args[0][0] == peer_info

    assert settings_manager.base == settings
    instance.error_shutdown("Ensure all threads and resources are cleaned up")


def test_create_instance_set_up_checkpointing(
        manager_location_argv, instance_argv, mmp_client, trigger_manager,
        no_resume_snapshot_manager, settings_manager, declared_ports):

    instance = Instance(declared_ports)

    elapsed_time, checkpoints, resume_path, snapshot_path = (
            mmp_client.get_checkpoint_info.return_value)

    trigger_manager.set_checkpoint_info.assert_called_with(elapsed_time, checkpoints)
    no_resume_snapshot_manager.prepare_resume.assert_called_with(
            resume_path, snapshot_path)
    assert settings_manager.overlay != no_resume_snapshot_manager.resume_overlay
    instance.error_shutdown("Ensure all threads and resources are cleaned up")


def test_create_instance_set_up_logging(
        manager_location_argv, instance_argv, settings_manager, declared_ports):

    def get_setting(instance, name, typ):
        return {
                'muscle_local_log_level': 'debug',
                'muscle_remote_log_level': 'error'}[str(name)]

    settings_manager.get_setting = get_setting

    root_logger = MagicMock()
    root_logger.isEnabledFor.return_value = False
    libmuscle_logger = MagicMock()
    ymmsl_logger = MagicMock()

    def get_logger(name=''):
        return {
                '': root_logger,
                'libmuscle': libmuscle_logger,
                'ymmsl': ymmsl_logger}[name]

    with patch('libmuscle.instance.logging.getLogger', get_logger):
        instance = Instance(declared_ports)

        assert instance._mmp_handler.level == logging.ERROR
        root_logger.setLevel.assert_called_with(logging.ERROR)
        libmuscle_logger.setLevel.assert_called_with(logging.DEBUG)
        ymmsl_logger.setLevel.assert_called_with(logging.DEBUG)
    instance.error_shutdown("Ensure all threads and resources are cleaned up")


def test_shutdown_instance(
        logger, instance, mmp_client, communicator, profiler):

    msg = 'Testing'
    num_profile_events = len(profiler.record_event.mock_calls)

    instance.error_shutdown(msg)

    logger.critical.assert_called_with(msg)
    communicator.shutdown.assert_called()

    mmp_client.deregister_instance.assert_called_once_with()
    mmp_client.close.assert_called_once_with()

    assert len(profiler.record_event.mock_calls) == num_profile_events + 1
    profiler.shutdown.assert_called_once_with()


def test_list_settings(instance, settings_manager):
    instance.list_settings()
    settings_manager.list_settings.assert_called_with(Ref('component'))


def test_get_setting(instance, settings_manager):
    settings_manager.get_setting.side_effect = None  # don't raise KeyError
    instance.get_setting('test', 'int')
    settings_manager.get_setting.assert_called_with(
            Ref('component'), Ref('test'), 'int')


def test_list_ports(instance, port_manager):
    port_manager.list_ports.assert_called_once_with()
    port_manager.list_ports.reset_mock()
    instance.list_ports()
    port_manager.list_ports.assert_called_once_with()


def test_is_connected(instance):
    assert instance.is_connected('in') is True
    assert instance.is_connected('not_connected') is False
    assert instance.is_connected('out_v') is True
    assert instance.is_connected('out_r') is True
    assert instance.is_connected('in_v') is True
    assert instance.is_connected('in_r') is True
    assert instance.is_connected('not_connected_v') is False
    assert instance.is_connected('out') is True


def test_is_vector_port(instance):
    assert instance.is_vector_port('in') is False
    assert instance.is_vector_port('not_connected') is False
    assert instance.is_vector_port('out_v') is True
    assert instance.is_vector_port('out_r') is True
    assert instance.is_vector_port('in_v') is True
    assert instance.is_vector_port('in_r') is True
    assert instance.is_vector_port('not_connected_v') is True
    assert instance.is_vector_port('out') is False


def test_is_resizable(instance):
    for port in ('in', 'not_connected', 'out_v', 'in_v', 'out'):
        assert instance.is_resizable(port) is False

    assert instance.is_resizable('out_r') is True
    assert instance.is_resizable('in_r') is True
    assert instance.is_resizable('not_connected_v') is True


def test_get_port_length(instance):
    assert instance.get_port_length('out_v') == 13
    assert instance.get_port_length('out_r') == 0
    assert instance.get_port_length('in_v') == 13
    assert instance.get_port_length('in_r') == 0


def test_set_port_length(instance, port_manager):
    instance.set_port_length('not_connected_v', 7)
    assert port_manager.get_port('not_connected_v').get_length() == 7


def test_reuse_set_overlay(
        instance, port_manager, mock_ports, communicator, settings_manager):
    port_manager.settings_in_connected.return_value = True
    mock_ports['in']._is_connected = False

    mock_msg = MagicMock()
    mock_msg.data = Settings({'s1': 1, 's2': 2})
    mock_msg.settings = Settings({'s0': 0})
    communicator.receive_message.return_value = mock_msg, 0.0

    instance.reuse_instance()

    communicator.receive_message.assert_called_with('muscle_settings_in')
    assert settings_manager.overlay['s0'] == 0
    assert settings_manager.overlay['s1'] == 1
    assert settings_manager.overlay['s2'] == 2


@pytest.mark.parametrize('closed_port', ['muscle_settings_in', 'in'])
def test_reuse_closed_port(instance, port_manager, communicator, closed_port):

    def receive_message(port, slot=None, default=None):
        mock_msg = MagicMock()
        if port == closed_port:
            mock_msg.data = ClosePort()
        else:
            mock_msg.data = Settings()
        return mock_msg, 0.0

    port_manager.settings_in_connected.return_value = True
    communicator.receive_message = receive_message

    assert instance.reuse_instance() is False


def test_reuse_f_init_vector_port(instance, port_manager, communicator):
    port_manager.get_port('in')._length = 10

    def receive_message(port, slot=None, default=None):
        mock_msg = MagicMock()
        mock_msg.data = Settings()
        return mock_msg, 0.0

    communicator.receive_message = receive_message

    assert instance.reuse_instance() is True


def test_reuse_no_f_init_ports(instance, connected_port_manager, communicator):
    connected_port_manager.list_ports.return_value = {}

    assert instance.reuse_instance() is True
    assert instance.reuse_instance() is False


def test_send_message(instance, settings_manager, communicator):
    port = 'out_v'
    msg = MagicMock()
    msg.settings = None
    slot = 3

    instance.send(port, msg, slot)

    communicator.send_message.assert_called_once()
    args = communicator.send_message.call_args[0]
    assert args[0] == port
    assert args[1].settings == settings_manager.overlay
    assert args[2] == slot


def test_send_on_invalid_port(instance):
    with pytest.raises(RuntimeError):
        instance.send('does_not_exist', MagicMock())


def test_send_after_resize(instance, message):
    with pytest.raises(RuntimeError):
        instance.send('out_r', message, 13)

    instance.set_port_length('out_r', 20)
    instance.send('out_r', message, 13)


def test_send_on_receiving_port(instance, message):
    with pytest.raises(RuntimeError):
        instance.send("in_v", message, 3)


def test_receive_on_invalid_port(instance):
    with pytest.raises(RuntimeError):
        instance.receive('does_not_exist')


def test_receive_on_sending_port(instance):
    with pytest.raises(RuntimeError):
        instance.receive("out_v", 3)


def test_receive_f_init(instance, port_manager, communicator):
    mock_msg = MagicMock()
    mock_msg.data = Settings()
    communicator.receive_message.return_value = mock_msg, 0.0

    instance.reuse_instance()

    msg = instance.receive('in')

    assert msg == mock_msg


def test_receive_default_f_init(instance):
    default_msg = MagicMock()

    msg = instance.receive('not_connected', default=default_msg)

    assert msg == default_msg


def test_receive_default(instance):
    default_msg = MagicMock()

    msg = instance.receive('not_connected_v', 42, default_msg)

    assert msg == default_msg


def test_receive_no_default(instance):
    with pytest.raises(RuntimeError):
        instance.receive('not_connected')

    with pytest.raises(RuntimeError):
        instance.receive('not_connected_v', 14)


def test_receive_inconsistent_settings(
        instance, settings_manager, port_manager, communicator):

    def receive_message(port, slot=None):
        mock_msg = MagicMock()
        if port == 'muscle_settings_in':
            mock_msg.data = Settings({'s1': 1})
            mock_msg.settings = Settings()
        else:
            mock_msg.data = None
            mock_msg.settings = Settings({'s0': 0})
        return mock_msg, 0.0

    communicator.receive_message.side_effect = receive_message

    port_manager.settings_in_connected.return_value = True

    with pytest.raises(RuntimeError):
        instance.reuse_instance()


def test_receive_with_settings(
        instance_dont_apply_overlay, settings_manager, communicator):

    mock_msg = MagicMock()
    mock_msg.settings = Settings({'s0': 0, 's1': 1})
    communicator.receive_message.return_value = mock_msg, 0.0

    instance_dont_apply_overlay.reuse_instance()
    msg = instance_dont_apply_overlay.receive('in')

    assert msg.settings['s0'] == 0
    assert msg.settings['s1'] == 1

    assert len(settings_manager.overlay) == 0


def test_receive_with_settings_default(
        instance_dont_apply_overlay, settings_manager, port_manager, communicator):

    port_manager.get_port('in')._is_connected = False

    instance_dont_apply_overlay.reuse_instance()

    default_msg = MagicMock()
    default_msg.settings = MagicMock()
    msg = instance_dont_apply_overlay.receive('in', default=default_msg)

    assert msg == default_msg
    assert msg.settings == default_msg.settings
    assert len(settings_manager.overlay) == 0


@pytest.mark.parametrize('flags, expectation', [
        (InstanceFlags(0), pytest.raises(RuntimeError)),
        (InstanceFlags.USES_CHECKPOINT_API, does_not_raise()),
        (InstanceFlags.KEEPS_NO_STATE_FOR_NEXT_USE, does_not_raise()),
        (InstanceFlags.STATE_NOT_REQUIRED_FOR_NEXT_USE, does_not_raise())])
def test_checkpoint_support(
        instance_argv, mmp_client, tmp_path, flags, expectation):

    checkpoint_info = (0.0, Checkpoints(at_end=True), None, tmp_path)
    mmp_client.get_checkpoint_info.return_value = checkpoint_info

    with expectation:
        instance = Instance(flags=flags)
        instance.error_shutdown("Ensure all threads and resources are cleaned up")
